#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.
import copy
import pickle
import time
from collections import defaultdict

import fasteners
import networkx as nx


from .targets import (sos_step, sos_targets, sos_variable, textMD5,
                      BaseTarget)
from .utils import ActivityNotifier, env, short_repr

from typing import Union
#
# DAG design:
#
# 1. Nodes are jobs to be completed currently they are the
#    same as steps.
#
# 2. Input, output and depedent files of nodes.
#
#    Each step has input, output and depdent files with their status.
#    The status of these files determines the status of ndoes.
#    There are several status for these targets
#
#    a. Target exists and without signature.
#       The files need to be processed and obtain its signature.
#
#    b. Target does not exists and with signature.
#       The files are assumed to exist for dependency check, but
#       are determined to be non-exist if they are needed for a
#       task.
#
#    c. Target does not exists and without signature.
#       The target need to be generated and obtain its signature.
#
#    d. Target is undetermined. A node with undetermined input
#       will depend on all its previous steps.
#
#  3. Status of node.
#
#    a. A node does not depend on any input file.
#       It can be executed immediately.
#
#    b. A node depends on undetermined input file.
#       It depend on all previous steps and can only be processed
#       when all its previous steps are implemented.
#
#    c. A node depends on a previous file.
#       It can only be executed when the previous step is executed.
#       This is the regular case.
#
#    d. A node depends on a file that does not exist.
#       This is problematic. An error will be raised.
#
# 4. Edges.
#
#    a. Edges determines dependencies between nodes. They are
#       generated depending on the status of the nodes.
#
#    b. Edges will be updated with the update of nodes.
#
# 5. Dynamic DAG.
#
#    a. DAG will consists of nodes and edges derived from
#       status of files.
#
#    b. Completion of a task or step will update the status of
#       files. All edges related to updated targets will be updated.
#
#    c. Addition of node is currently not considered, but should
#       be allowed.
#


class SoS_Node(object):
    def __init__(self, step_uuid: str, node_name: str, wf_index: Union[int, None], node_index: Union[int, None], input_targets: sos_targets, depends_targets: sos_targets,
                 output_targets: sos_targets, context: dict) -> None:
        self._step_uuid = step_uuid
        self._node_id = node_name
        self._wf_index = wf_index
        self._node_index = node_index
        self._input_targets = input_targets
        self._depends_targets = depends_targets
        self._output_targets = output_targets
        # env.logger.error('Note {}: Input: {} Depends: {} Output: {}'.format(self._node_id, self._input_targets,
        #      self._depends_targets,  self._output_targets))
        self._context = {} if context is None else copy.deepcopy(context)
        self._status = None
        # unique ID to avoid add duplicate nodes ...
        self._node_uuid = textMD5(pickle.dumps((step_uuid, node_name, node_index,
                                                input_targets, depends_targets, output_targets,
                                                [] if context is None else [(k, sorted(list(context[k])) if isinstance(context[k], set) else context[k])
                                                                            for k in sorted(context.keys())])))

    def __repr__(self):
        return self._node_id

    def show(self):
        print(
            f'{self._node_id} ({self._node_index}, {self._status}): input {short_repr(self._input_targets)}, depends {short_repr(self._depends_targets)}, output {short_repr(self._output_targets)}, context {self._context}')


class SoS_DAG(nx.DiGraph):
    def __init__(self, *args, **kwargs):
        nx.DiGraph.__init__(self, *args, **kwargs)
        # all_input
        self._all_depends_files = defaultdict(list)
        self._all_output_files = defaultdict(list)
        # index of mini
        self._forward_workflow_id = 0
        # if dag has been changed
        self._dirty = True

    def mark_dirty(self, dirty=True):
        self._dirty = dirty

    def dirty(self):
        return self._dirty

    def new_forward_workflow(self):
        self._forward_workflow_id += 1

    def num_nodes(self):
        return nx.number_of_nodes(self)

    def add_step(self, step_uuid, node_name, node_index, input_targets: sos_targets, depends_targets: sos_targets,
                 output_targets: sos_targets, context: dict={}):
        if 'DAG' in env.config['SOS_DEBUG']:
            env.log_to_file('DAG', f'add step {node_name}')
        node = SoS_Node(step_uuid, node_name, None if node_index is None else self._forward_workflow_id, node_index, input_targets, depends_targets,
                        output_targets, context)
        if node._node_uuid in [x._node_uuid for x in self.nodes()]:
            return
        # adding a step would add a sos_step target to met the depends on sos_step
        # requirement of some steps.
        self._all_output_files[sos_step(node_name.split(' ')[0])].append(node)

        for x in input_targets:
            if node not in self._all_depends_files[x]:
                self._all_depends_files[x].append(node)
        for x in depends_targets:
            if node not in self._all_depends_files[x]:
                self._all_depends_files[x].append(node)
        for x in output_targets:
            if node not in self._all_output_files[x]:
                self._all_output_files[x].append(node)
        if context is not None:
            for x in context['__changed_vars__']:
                if node not in self._all_output_files[sos_variable(x)]:
                    self._all_output_files[sos_variable(x)].append(node)
        self.add_node(node)

    def update_step(self, node, input_targets: sos_targets, output_targets: sos_targets, depends_targets: sos_targets):
        for x in input_targets:
            if node not in self._all_depends_files[x]:
                self._all_depends_files[x].append(node)
        for x in depends_targets:
            if node not in self._all_depends_files[x]:
                self._all_depends_files[x].append(node)
        for x in output_targets:
            if node not in self._all_output_files[x]:
                self._all_output_files[x].append(node)

    def find_executable(self):
        '''Find an executable node, which means nodes that has not been completed
        and has no input dependency.'''
        if 'DAG' in env.config['SOS_DEBUG']:
            env.log_to_file('DAG', 'find_executable')
        for node in self.nodes():
            # if it has not been executed
            if node._status is None:
                with_dependency = False
                for edge in self.in_edges(node):
                    if edge[0]._status != 'completed':
                        with_dependency = True
                        break
                if not with_dependency:
                    return node
        # if no node could be found, let use try pending ones
        pending_jobs = [x for x in self.nodes() if x._status ==
                        'signature_pending']
        if pending_jobs:
            try:
                notifier = ActivityNotifier(
                    f'Waiting for {len(pending_jobs)} pending job{"s: e.g." if len(pending_jobs) > 1 else ":"} output {short_repr(pending_jobs[0]._signature[0])} with signature file {pending_jobs[0]._signature[1] + "_"}. You can manually remove this lock file if you are certain that no other process is working on the output.')
                while True:
                    for node in pending_jobs:
                        # if it has not been executed
                        lock = fasteners.InterProcessLock(
                            node._signature[1] + '_')
                        if lock.acquire(blocking=False):
                            lock.release()
                            node._status = None
                            return node
                    time.sleep(0.1)
            except Exception as e:
                env.logger.error(e)
            finally:
                notifier.stop()
        return None

    def node_by_id(self, node_uuid):
        for node in self.nodes():
            if node._node_uuid == node_uuid:
                return node
        raise RuntimeError(f'Failed to locate node with UUID {node_uuid}')

    def show_nodes(self):
        for node in self.nodes():
            node.show()
        for edge in self.edges():
            print(edge)

    def circular_dependencies(self):
        if 'DAG' in env.config['SOS_DEBUG']:
            env.log_to_file('DAG', 'check circular')
        try:
            return nx.find_cycle(self)
        except Exception:
            # if there is no cycle, an exception is given
            return []

    def steps_depending_on(self, target: BaseTarget, workflow):
        if target in self._all_depends_files:
            return ' requested by ' + ', '.join(set([workflow.section_by_id(x._step_uuid).step_name() for x in self._all_depends_files[target]]))
        return ''

    def pending(self):
        return [x for x in self.nodes() if x._status == 'failed'], [x for x in self.nodes() if x._status is None]

    def dangling(self, targets: sos_targets):
        '''returns
        1. missing targets, which are missing from the DAG or from the provided targets
        2. existing targets of provided target list, not in DAG
        '''
        existing = []
        missing = []
        if env.config['trace_existing']:
            for x in self._all_depends_files.keys():
                if x not in self._all_output_files:
                    if x.target_exists():
                        existing.append(x)
                    else:
                        missing.append(x)
        else:
            missing = [x for x in self._all_depends_files.keys() if x not in self._all_output_files and not x.target_exists()]
        for x in targets:
            if x not in self._all_output_files:
                if x.target_exists('target'):
                    existing.append(x)
                else:
                    missing.append(x)
        return missing, existing

    def regenerate_target(self, target: BaseTarget):
        if target in self._all_output_files:
            for node in self._all_output_files[target]:
                if node._status == 'completed':
                    if isinstance(target, sos_step):
                        raise RuntimeError(
                            f'Completed target {target} is being re-executed. Please report this bug to SoS developers.')
                    else:
                        env.logger.info(
                            f'Re-running {node._node_id} to generate {target}')
                        node._status = None
            return True
        else:
            return False

    def subgraph_from(self, targets: sos_targets):
        '''Trim DAG to keep only nodes that produce targets'''
        if 'DAG' in env.config['SOS_DEBUG']:
            env.log_to_file('DAG', 'create subgraph')
        # first, find all nodes with targets
        subnodes = []
        for node in self.nodes():
            if node._output_targets.valid() and any(x in node._output_targets for x in targets):
                subnodes.append(node)
        #
        ancestors = set()
        for node in subnodes:
            ancestors |= nx.ancestors(self, node)
        return SoS_DAG(nx.subgraph(self, subnodes + list(ancestors)))

    def build(self):
        '''Connect nodes according to status of targets'''
        # right now we do not worry about status of nodes
        # connecting the output to the input of other nodes
        #
        # NOTE: This is implemented in the least efficient way just for
        # testing. It has to be re-implemented.
        #
        # refer to http://stackoverflow.com/questions/33494376/networkx-add-edges-to-graph-from-node-attributes
        #
        # several cases triggers dependency.
        if 'DAG' in env.config['SOS_DEBUG']:
            env.log_to_file('DAG', 'build DAG')
        for wf in range(self._forward_workflow_id + 1):
            indexed = [x for x in self.nodes() if x._wf_index == wf]
            indexed.sort(key=lambda x: x._node_index)

            for idx, node in enumerate(indexed):
                # 1. if a node changes context (using option alias), all later steps
                # has to rely on it.
                if node._context['__changed_vars__']:
                    for later_node in indexed[idx + 1:]:
                        if node._context['__changed_vars__'] & (later_node._context['__signature_vars__'] | later_node._context['__environ_vars__']):
                            self.add_edge(node, later_node)

                # 2. if the input of a step is undetermined, it has to be executed
                # after all its previous steps.
                if not node._input_targets.valid() and idx > 0:
                    # if there is some input specified, it does not use default
                    # input, so the relationship can be further looked before
                    if node._input_targets.undetermined():
                        # if the input is dynamic, has to rely on previous step...
                        if 'dynamic' in node._context['__environ_vars__']:
                            self.add_edge(indexed[idx - 1], node)
                        else:
                            # otherwise let us look back.
                            for prev_node in indexed[idx - 1::-1]:
                                if node._context['__environ_vars__'] & prev_node._context['__changed_vars__']:
                                    self.add_edge(prev_node, node)
                    else:
                        self.add_edge(indexed[idx - 1], node)
        #
        # 3. if the input of a step depends on the output of another step
        for target, in_node in self._all_depends_files.items():
            if target not in self._all_output_files:
                continue
            # it is possible that multiple nodes satisfy the same target
            out_node = self._all_output_files[target]
            for i in in_node:
                for j in out_node:
                    if j != i:
                        self.add_edge(j, i)
        self.mark_dirty()

    def save(self, dest=None):
        if not dest:
            return

        if 'DAG' in env.config['SOS_DEBUG']:
            env.log_to_file('DAG', 'save DAG')
        if not hasattr(self, 'dag_count'):
            self.last_dag = None
        #
        for x in self.nodes():
            if x._status is None:
                pass
            elif x._status == 'running':
                self.add_node(x, color='green')
            elif x._status == 'failed':
                self.add_node(x, color='red')
            elif 'pending' in x._status:
                self.add_node(x, color='yellow')
            elif x._status == 'completed':
                self.add_node(x, color='blue')
            elif x._status is not None:
                env.logger.warning(f'Unmarked step status {x._status}')

        try:
            out = nx.drawing.nx_pydot.to_pydot(self).to_string()
        except Exception as e:
            env.logger.warning(f'Failed to call to_pydot: {e}')

        if self.last_dag == out:
            return
        else:
            self.last_dag = out
        # output file name
        if hasattr(dest, 'write'):
            dest.write(out)
        else:
            with open(dest, 'a' if self.last_dag else 'w') as dfile:
                dfile.write(out)



#!/usr/bin/env python
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import argparse
import base64
import io
import pkg_resources
from sos.utils import dehtml, env, dot_to_gif, linecount_of_file


def get_previewers():
    # Note: data is zest.releaser specific: we want to pass
    # something to the plugin
    group = 'sos_previewers'
    result = []
    for entrypoint in pkg_resources.iter_entry_points(group=group):
        # if ':' in entry point name, it should be a function
        try:
            name, priority = entrypoint.name.split(',', 1)
            priority = int(priority)
        except Exception as e:
            env.logger.warning(
                f'Ignore incorrect previewer entry point {entrypoint}: {e}')
            continue
        # If name points to a function in a module. Let us try to import the module
        if ':' in name:
            import importlib
            try:
                mod, func = name.split(':')
                imported = importlib.import_module(mod)
                result.append((getattr(imported, func), entrypoint, priority))
            except ImportError:
                env.logger.warning(f'Failed to load function {mod}:{func}')
        else:
            result.append((name, entrypoint, priority))
    #
    result.sort(key=lambda x: -x[2])
    # we put string infront of functions so that filename matching will be used before
    # content matching. For example, the xlsx file is actually a zip file so it could be
    # previewed as a zip file first.
    return [x for x in result if isinstance(x[0], str) and x[0] != '*'] + \
        [x for x in result if not isinstance(x[0], str)] + \
        [x for x in result if x[0] == '*']


def _preview_img_parser():
    parser = argparse.ArgumentParser(prog='%preview *.pdf')
    parser.add_argument('--width',
        help='Width of the previewed image, can be in any HTML units such as px, em.')
    parser.add_argument('--height',
        help='Height of the previewed image, can be in any HTML units such as px, em.')
    parser.error = lambda msg: env.logger.warning(msg)
    return parser

def preview_img(filename, kernel=None, style=None):
    with open(filename, 'rb') as f:
        image = f.read()
    import imghdr
    image_type = imghdr.what(None, image)
    image_data = base64.b64encode(image).decode('ascii')

    args = None
    meta = {}
    if style is not None and 'options' in style:
        parser = _preview_img_parser()
        try:
            args = parser.parse_args(style['options'])
            meta.update({'image/png':
                dict(([['width', args.width]] if args.width else []) +
                    ([['height', args.height]] if args.height else []))})
        except SystemExit:
            return

    if image_type != 'png':
        try:
            if image_type == 'gif':
                return {'image/png': image_data}, meta
            else:
                from wand.image import Image
                img = Image(filename=filename)
                return {'image/' + image_type: image_data,
                        'image/png': base64.b64encode(img._repr_png_()).decode('ascii')}, meta
        except Exception:
            return {'image/' + image_type: image_data}, meta
    else:
        return {'image/' + image_type: image_data}, meta


def preview_svg(filename, kernel=None, style=None):
    with open(filename, 'r') as f:
        image_data = f.read()
    return {'image/svg+xml': image_data}


def _preview_pdf_parser():
    parser = argparse.ArgumentParser(prog='%preview *.pdf')
    parser.add_argument('--pages', nargs='+', type=int,
                        help='Pages of the PDF to preview.')
    parser.add_argument('--width',
        help='Width of the previewed image, can be in any HTML units such as px, em.')
    parser.add_argument('--height',
        help='Height of the previewed image, can be in any HTML units such as px, em.')
    parser.add_argument('--dpi', type=int, default=150,
        help='resolution of the converted png preview')
    parser.error = lambda msg: env.logger.warning(msg)
    return parser


def preview_pdf(filename, kernel=None, style=None):
    use_png = False
    warn = kernel.warn if kernel is not None else env.logger.warning
    if style is not None and 'style' in style:
        if style['style'] != 'png':
            if style['style'] is not None:
                warn(
                    f'Option --style of PDF preview only accept parameter png: {style["style"]} provided')
        else:
            use_png = True
    args = None
    if style is not None and 'options' in style:
        parser = _preview_pdf_parser()
        try:
            args = parser.parse_args(style['options'])
        except SystemExit:
            return
    meta = {}
    embed_options = ''
    if args and (args.width or args.height):
        meta.update({'image/png':
            dict(([['width', args.width]] if args.width else []) +
                ([['height', args.height]] if args.height else []))})
        embed_options += (f'width="{args.width}" ' if args.width else ' ') + \
            (f'height="{args.height}" ' if args.height else ' ')
    if use_png:
        try:
            # this import will fail even if wand is installed
            # if imagemagick is not installed properly.
            from wand.image import Image
            img = Image(filename=filename, resolution=args.dpi)

            if img.width == 0 or img.height == 0:
                raise ValueError('Image appears to have zero width or height')
            nPages = len(img.sequence)
            pages = list(range(nPages))

            if args and args.pages is not None:
                pages = [x - 1 for x in args.pages]
                for p in pages:
                    if p >= nPages:
                        warn(
                            f'Page {p} out of range of the pdf file ({nPages} pages)')
                        pages = list(range(nPages))
                        break
            # single page PDF
            if len(pages) == 1 and pages[0] == 0:
                return {
                    'image/png': base64.b64encode(img._repr_png_()).decode('ascii')}, meta
            elif len(pages) == 1:
                # if only one page
                return {
                    'image/png': base64.b64encode(Image(img.sequence[pages[0]])._repr_png_()).decode('ascii')}, meta
            else:
                widths = [img.sequence[p].width for p in pages]
                heights = [img.sequence[p].height for p in pages]
                image = Image(width=max(widths), height=sum(heights))
                for i, p in enumerate(pages):
                    image.composite(
                        img.sequence[p],
                        top=sum(heights[:i]),
                        left=0
                    )
                image.format = 'png'
                with io.BytesIO() as out:
                    image.save(file=out)
                    img_data = out.getvalue()
                return {
                    'image/png': base64.b64encode(img_data).decode('ascii')}, meta
        except Exception as e:
            warn(e)
            return {'text/html':
                    f'<embed src="{filename}" {embed_options} type="application/pdf" />'}
    else:
        # by default use iframe, because PDF figure can have multiple pages (#693)
        return {'text/html':
                f'<embed src="{filename}" {embed_options} type="application/pdf" />'}


def preview_html(filename, kernel=None, style=None):
    with open(filename) as html:
        content = html.read()
    return {'text/html': content,
            'text/plain': dehtml(content)}


def _get_txt_parser():
    parser = argparse.ArgumentParser(prog='%preview *.txt')
    parser.add_argument('-l', '--limit', type=int, default=5, help='''Maximum number
        of lines to preview.''')
    parser.error = lambda msg: env.logger.warning(msg)
    return parser


def preview_txt(filename, kernel=None, style=None):
    if style is not None and 'options' in style:
        parser = _get_txt_parser()
        try:
            args = parser.parse_args(style['options'])
        except SystemExit:
            return
        limit = args.limit
    else:
        limit = 5

    nlines = linecount_of_file(filename)
    content = 'HINT: {} line{}{}\n'.format(nlines, 's' if nlines > 1 else '',
                                          f' ({limit} displayed, see --limit)' if nlines > limit else '')

    with open(filename, 'r') as fin:
        if limit < 0:
            content += fin.read()
        else:
            for _ in range(limit):
                content += fin.readline()
    return content


def preview_csv(filename, kernel=None, style=None):
    import pandas
    from .visualize import Visualizer
    data = pandas.read_csv(filename)
    return Visualizer(kernel, style).preview(data)


def preview_xls(filename, kernel=None, style=None):
    import pandas
    from .visualize import Visualizer
    data = pandas.read_excel(filename)
    return Visualizer(kernel, style).preview(data)


def preview_zip(filename, kernel=None, style=None):
    import zipfile
    names = zipfile.ZipFile(filename).namelist()
    return f'{len(names)} files\n' + '\n'.join(names[:5]) + ('\n...' if len(names) > 5 else '')


def preview_tar(filename, kernel=None, style=None):
    import tarfile
    with tarfile.open(filename, 'r:*') as tar:
        # only extract files
        names = [x.name for x in tar.getmembers() if x.isfile()]
    return f'{len(names)} files\n' + '\n'.join(names[:5]) + ('\n...' if len(names) > 5 else '')


def preview_gz(filename, kernel=None, style=None):
    import tarfile
    # .tar.gz
    if tarfile.is_tarfile(filename):
        return preview_tar(filename, kernel, style)
    import gzip
    content = b''
    with gzip.open(filename, 'rb') as fin:
        for _ in range(5):
            content += fin.readline()
    try:
        return content.decode()[:2000]
    except Exception:
        return 'binary data'


def preview_md(filename, kernel=None, style=None):
    import markdown
    with open(filename) as fin:
        text = fin.read()
    html = markdown.markdown(text)
    return {'text/html': html, 'text/plain': text}


def preview_dot(filename, kernel=None, style=None):
    data = dot_to_gif(
        filename, warn=kernel.warn if kernel else env.logger.warning)
    # according to https://github.com/ipython/ipython/issues/10045
    # I have to use 'image/png' instead of 'image/gif' to get the gif displayed.
    return {'image/png': data}



#!/usr/bin/env python
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import shutil
import unittest

from sos.parser import SoS_Script
from sos.targets import file_target
from sos.utils import env
from sos.workflow_executor import Base_Executor


class TestActions(unittest.TestCase):
    def setUp(self):
        env.reset()
        self.temp_files = []

    def tearDown(self):
        for f in self.temp_files:
            file_target(f).unlink()

    def touch(self, files):
        '''create temporary files'''
        if isinstance(files, str):
            files = [files]
        #
        for f in files:
            with open(f, 'w') as tmp:
                tmp.write('test')
        #
        self.temp_files.extend(files)

    @unittest.skipIf(not shutil.which('python2.7'), 'Skip test because of no python2.7 installation')
    def testPython2(self):

        script = SoS_Script(r'''
[0]
python2: expand='${ }'
a = {'1', '2'}
print a
''')
        wf = script.workflow()
        Base_Executor(wf).run()


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import os
import shutil
import subprocess
import sys
import tempfile

from sos.actions import SoS_Action, SoS_ExecuteScript, collect_input
from sos.eval import interpolate
from sos.targets import sos_targets
from sos.utils import env

from .targets_r import R_library


@SoS_Action(acceptable_args=['script', 'args'], default_args={'default_env': {'R_DEFAULT_PACKAGES': 'datasets,methods,utils,stats,grDevices,graphics'}})
def R(script, args='', **kwargs):
    '''Execute specified script with command Rscript, with default options
    "--default-packages=datasets,methods,utils,stats,grDevices,graphics". This action accepts
    common action arguments such as input, active, workdir, docker_image and args.
    In particular, content of one or more files  specified by option input would be
    prepended before the specified script.
    '''
    # > getOption('defaultPackages')
    # [1] "datasets"  "utils"     "grDevices" "graphics"  "stats"     "methods"
    return SoS_ExecuteScript(
        script, 'Rscript', '.R', args).run(**kwargs)


@SoS_Action(acceptable_args=['script', 'args'])
def Rmarkdown(script=None, input=None, output=None, args='{input:r}, output_file={output:ar}', **kwargs):
    '''Convert input file to output using Rmarkdown

    The input can be specified in three ways:

    1. instant script, which is assumed to be in md format

    Rmarkdown:   output='report.html'
      script

    2. one or more input files. The format is determined by extension of input file

    Rmarkdown(input, output='report.html')

    3. input file specified by command line option `-r` .
    Rmarkdown(output='report.html')

    If no output is specified, it is assumed to be in html format
    and is written to standard output.

    You can specify more options using the args parameter of the action. The default value
    of args is `${input!r} --output ${output!ar}'
    '''
    if not R_library('rmarkdown').target_exists():
        raise RuntimeError('Library rmarkdown does not exist')

    input = sos_targets(collect_input(script, input))

    output = sos_targets(output)
    if len(output) == 0:
        write_to_stdout = True
        output = sos_targets(tempfile.NamedTemporaryFile(
            mode='w+t', suffix='.html', delete=False).name)
    else:
        write_to_stdout = False
    #
    ret = 1
    try:
        #   render(input, output_format = NULL, output_file = NULL, output_dir = NULL,
        #        output_options = NULL, intermediates_dir = NULL,
        #        runtime = c("auto", "static", "shiny"),
        #        clean = TRUE, params = NULL, knit_meta = NULL, envir = parent.frame(),
        #        run_Rmarkdown = TRUE, quiet = FALSE, encoding = getOption("encoding"))
        cmd = interpolate(f'Rscript -e "rmarkdown::render({args})"',
                          {'input': input, 'output': output})
        if 'ACTION' in env.config['SOS_DEBUG']:
            env.log_to_file('ACTION', f'Running command "{cmd}"')
        if env.config['run_mode'] == 'interactive':
            # need to catch output and send to python output, which will in trun be hijacked by SoS notebook
            p = subprocess.Popen(cmd, shell=True, stderr=subprocess.PIPE, stdout=subprocess.PIPE)
            #pid = p.pid
            out, err = p.communicate()
            sys.stdout.write(out.decode())
            sys.stderr.write(err.decode())
            ret = p.returncode
        else:
            p = subprocess.Popen(cmd, shell=True)
            #pid = p.pid
            ret = p.wait()
    except Exception as e:
        env.logger.error(e)
    if ret != 0:
        temp_file = os.path.join('.sos', f'{"Rmarkdown"}_{os.getpid()}.md')
        shutil.copyfile(str(input), temp_file)
        cmd = interpolate(f'Rscript -e "rmarkdown::render({args})"',
                          {'input': input, 'output': sos_targets(temp_file)})
        raise RuntimeError(
            f'Failed to execute script. Please use command \n"{cmd}"\nunder {os.getcwd()} to test it.')
    if write_to_stdout:
        with open(str(output[0])) as out:
            sys.stdout.write(out.read())
    else:
        env.logger.info(f'Report saved to {output}')
